# -*- coding: utf-8 -*-

from os.path import exists, isdir, join
try:
    from cPickle import load
except ImportError:
    from pickle import load
from django.core.urlresolvers import reverse as urlreverse
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.views import static
from models import Domain


# decorators
def force_trailing_slash(func):
    """Forces view to redirect slash-suffixed url.
    """
    def _wrapped(request, *args, **kwargs):
        if request.path and not request.path.endswith('/'):
            return HttpResponseRedirect(request.path+'/')
        return func(request, *args, **kwargs)
    return _wrapped


def bind_to_domain(standalone_func):
    """Call standalone func with 'document_root' parameter, which is
    resolved by domain_name parameter of resulting function.
    """
    def _wrapped(*args, **kwargs):
        domain = get_object_or_404(Domain, name=kwargs.pop('domain_name'))
        kwargs['document_root'] = domain.document_root
        extra_context = kwargs.pop('extra_context', {})
        extra_context.update({'domain': domain})
        kwargs['extra_context'] = extra_context
        return standalone_func(*args, **kwargs)
    return _wrapped
    

# path utility
def normalize_filename(filename, **kwargs):
    if (kwargs.get('strip_index', False) and
        filename.endswith('index')):
        filename = filename[:0-len('index')]
    return filename


# standalone handlers
@force_trailing_slash
def genindex_standalone(
    request, document_root='', template_name='lionet/genindex.html',
    serve_static=False, context_instance=None, extra_context=None):
    """Serves genindex under the directory specifie by 'document_root'.
    """
    params = {}
    # load genindex.
    try:
        genindex_path = join(document_root, 'genindex.fpickle')
        params['genindex'] = genindex = load(open(genindex_path, 'rb'))
        params['index_entries'] = genindex['genindexentries']
    except IOError:
        params['index_entries'] = []
    params.update(extra_context or {})
    return render_to_response(
        template_name, params, context_instance or RequestContext(request))


@force_trailing_slash
def modindex_standalone(
    request, document_root='', template_name='lionet/modindex.html',
    serve_static=False, context_instance=None, extra_context=None):
    params = {}
    """Serves modindex under the directory specifie by 'document_root'.
    """
    # load modindex.
    try:
        modindex_path = join(document_root, 'modindex.fpickle')
        params['modindex'] = modindex = load(open(modindex_path, 'rb'))
        params['index_entries'] = modindex['modindexentries']
    except IOError:
        params['index_entries'] = []
    params.update(extra_context or {})
    return render_to_response(
        template_name, params, context_instance or RequestContext(request))


@force_trailing_slash
def search_standalone(
    request, document_root='', template_name='lionet/search.html',
    serve_static=False, context_instance=None, extra_context=None,
    remove_index_from_pathname=True):
    """Searches keywords from searchindex.
    """
    params = {}
    # load searchindex.
    try:
        searchindex_path = join(document_root, 'searchindex.pickle')
        params['searchindex'] = searchindex = load(open(searchindex_path, 'rb'))
    except IOError:
        raise Http404
    # find match for query.
    query = request.GET.get('q', '')
    params['query'] = query
    hits = set()
    # split into keywords
    for kw in filter(len, query.strip().split()):
        # try substring match, starting from 1 chars.
        for i in range(len(kw)):
            subhits = searchindex['terms'].get(kw[:i+1], None)
            if subhits:
                if not isinstance(subhits, list):
                    subhits = [subhits]
                # hits are always merged (OR-behaviour).
                hits |= set(subhits)
                break
    # sort results by title.
    params['hits'] = sorted(
        (dict(title=searchindex['titles'][i],
              docname=normalize_filename(searchindex['filenames'][i],
                                         strip_index=remove_index_from_pathname))
         for i in hits),
        key=lambda e: e['title'])
    params.update(extra_context or {})
    return render_to_response(
        template_name, params, context_instance or RequestContext(request))


@force_trailing_slash
def serve_standalone(
    request, path, document_root='', template_name='lionet/default.html',
    serve_static=False, context_instance=None, extra_context=None):
    """Serves content under the directory specified by 'document_root'.
    """
    params = {}
    # load environment
    try:
        env_path = join(document_root, 'environment.pickle')
        environment = load(open(env_path, 'rb'))
    except IOError:
        raise Http404
    params['environment'] = environment
    # serving static files by django.views.static.serve,
    # if serve_static is True.
    if serve_static:
        if path.startswith('_static') or path.startswith('_source'):
            return static.serve(
                request, path, document_root, show_indexes=True)
    # master doc treatment
    doc_path = join(document_root, path.rstrip('/'))
    if path.strip('/')=='':
        masterdocname = environment.config.config_values.get('master_doc')[0]
        if exists(doc_path+'/'+masterdocname+'.fpickle'):
            return HttpResponseRedirect(masterdocname)
        # try to implicit masterdoc
        elif exists(doc_path+'/'+'index.fpickle'):
            doc_path += '/index'
        else:
            raise Http404
    # load pickled document.
    if isdir(doc_path):
        # if doc_path points to a directory, search for index.fpickle.
        doc_path += '/index.fpickle'
    else:
        # else, just suffix '.fpickle'.
        doc_path += '.fpickle'
    # load document and serve.
    try:
        document = load(open(doc_path, 'rb'))
    except IOError:
        raise Http404
    params['document'] = document
    params.update(extra_context or {})
    return render_to_response(
        template_name, params,
        context_instance or RequestContext(request))


# multiple handlers
def index_multiple(
    request, order_by='name', template_name='lionet/index.html',
    context_instance=None, extra_context=None, **kwargs):
    """Shows a calalog of domains.
    """
    params = {}
    params['domains'] = Domain.objects.all().order_by(order_by)
    return render_to_response(
        template_name, params,
        context_instance or RequestContext(request))
    

genindex_multiple = bind_to_domain(genindex_standalone)
modindex_multiple = bind_to_domain(modindex_standalone)
search_multiple = bind_to_domain(search_standalone)
serve_multiple = bind_to_domain(serve_standalone)


# 'full' handlers
@force_trailing_slash
def genindex_full(
    request, template_name='lionet/genindex.html',
    serve_static=False, context_instance=None, extra_context=None):
    """Generates cross-project genindex.
    """
    params = {}
    index_entries_dict = {}
    for domain in Domain.objects.all():
        domain_root_url = urlreverse('genindex_multiple', args=(domain.name,))
        # load genindex for each domain
        try:
            part_genindex_path = join(domain.document_root, 'genindex.fpickle')
            part_genindex = load(open(part_genindex_path, 'rb'))
            part_genindexentries = part_genindex['genindexentries']
            # edit entry of genindex: prefixing urls with domain_root_url
            # as well as appending domain.label to the entry.
            for k, v in part_genindexentries:
                index_entries_dict.setdefault(k, []).extend(
                    [(e,
                      ([domain_root_url+el for el in elinks],
                       [(subentry, [domain_root_url+sl for sl in sublinks])
                        for subentry, sublinks in subentries]),
                     domain.label)
                     for e, (elinks, (subentries)) in v])
        except IOError:
            pass
    # sort aggregated entries
    params['index_entries'] = sorted(
        (label, sorted(entries)) for label, entries in index_entries_dict.items())
    params.update(extra_context or {})
    return render_to_response(
        template_name, params, context_instance or RequestContext(request))


@force_trailing_slash
def modindex_full(
    request, template_name='lionet/modindex.html',
    serve_static=False, context_instance=None, extra_context=None):
    """Generates cross-project modindex.
    """
    params = {}
    index_entries = []
    for domain in Domain.objects.all():
        domain_root_url = urlreverse('modindex_multiple', args=(domain.name,))
        # load modindex for each domain
        try:
            part_modindex_path = join(domain.document_root, 'modindex.fpickle')
            part_modindex = load(open(part_modindex_path, 'rb'))
            part_modindexentries = part_modindex['modindexentries']
            # edit entry of modindex: prefixing urls with domain_root_url
            # as well as appending domain.label to the entry.
            for entry in part_modindexentries:
                if entry[0] == '':
                    if entry in index_entries:
                        continue
                    else:
                        index_entries.append(entry+[domain.label])
                else:
                    entry[4] = domain_root_url + entry[4]
                    index_entries.append(entry+[domain.label])
        except IOError:
            pass
    # sort aggregated entries, falling back from title to url.
    params['index_entries'] = sorted(
        index_entries,
        cmp=lambda a, b: cmp(str(a[0] or a[4]).lower(), str(b[0] or b[4]).lower()))
    params.update(extra_context or {})
    return render_to_response(
        template_name, params, context_instance or RequestContext(request))


@force_trailing_slash
def search_full(
    request, template_name='lionet/search.html',
    serve_static=False, context_instance=None, extra_context=None,
    remove_index_from_pathname=True):
    """Performs cross-project serach.
    """
    params = {}
    query = request.GET.get('q', '')
    params['query'] = query
    merged_hits = []
    for domain in Domain.objects.all():
        domain_root_url = urlreverse('modindex_multiple', args=(domain.name,))
        # load searchindex for each domain
        try:
            searchindex_path = join(domain.document_root, 'searchindex.pickle')
            searchindex = load(open(searchindex_path, 'rb'))
            hits = set()
            # edit entry of modindex: prefixing urls with domain_root_url
            for kw in filter(len, query.strip().split()):
                for i in range(len(kw)):
                    subhits = searchindex['terms'].get(kw[:i+1], None)
                    if subhits:
                        if not isinstance(subhits, list):
                            subhits = [subhits]
                        hits |= set(subhits)
                        break
            # merge hists, appending domain.label to the entry.
            merged_hits.extend(
                (dict(title=searchindex['titles'][i],
                      docname=normalize_filename(searchindex['filenames'][i],
                                         strip_index=remove_index_from_pathname))
                 for i in hits))
        except IOError:
            pass
    # sort hists by title.
    params['hits'] = sorted(sorted(merged_hits, key=lambda e: e['title']),
                            key=lambda e: e['domain'])
    params.update(extra_context or {})
    return render_to_response(
        template_name, params, context_instance or RequestContext(request))
